/* (c) by G. Caronni in 1995			19.11.95 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "cert_defs.h"
#include "cert_udp.h"
#include "cert_db.h"
#include "cert_int.h"
#ifdef UBS_CHIPCARD
#include "cert_ubs.h"
#endif
#include "md5.h"

#ifdef USE_SUNCERT
#include "suncert/enlink.h"

#ifdef __GNUC__
#ident "$Id: cert_udp.c,v 1.10 1996/06/07 11:32:57 skip Exp $"
#else
static char rcsid[] = "$Id: cert_udp.c,v 1.10 1996/06/07 11:32:57 skip Exp $";
#endif
#endif


int send_socket;
int recv_socket;


int udp_init(void)
{
    struct sockaddr_in addr;

    if ((send_socket = socket(PF_INET,SOCK_DGRAM,0)) < 0) return -1;
    memset(&addr,0,sizeof (addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(SEND_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(send_socket,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
        close(send_socket);
        perror("CERT udp_init bind send_socket");
        return -1;
    }
    if ((recv_socket = socket(PF_INET,SOCK_DGRAM,0)) < 0) {
        close(send_socket);
	return -1;
    }
    memset(&addr,0,sizeof (addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(RECV_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(recv_socket,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
        perror("CERT udp_init bind recv_socket");
        close(send_socket);
        close(recv_socket);
        return -1;
    }
    return 0;
}

void udp_exit(void)
{
    close(send_socket);
    close(recv_socket);
}

static dh_data *add_cert(u_char *buf, struct in_addr from)
{
    dh_data *res=NULL;
    int len, nsid, ctype, clen;
    u_char *mkid;

    u_char *t, *base;
    u_char *hex;
    int l,i;
    struct MD5Context ctxt;

    void *cdata;

/* current format is Name Type 
 *			SKIP=1
 *		     Name Length (8bit)
 *		     Name 
 *			NSID (8bit), MKID
 *		     Cert Type 
 *			DH Public = 4	OR
 *			X.509 = 1
 *		     Cert Length (16 bit)
 *		     Cert
 */

    base=buf;

    if (*buf++ != 1) {
        printf("CERT add_cert name type is not SKIP.\n");
	return NULL;
    }
    len= *buf++;
    nsid= *buf++;
    if (len-1 != mkid_size[nsid]) {
        printf("CERT add_cert mkid size not correct.\n");
	return NULL;
    }
    mkid=buf;
    buf += len-1;

    ctype= *buf++;
    clen=*buf++<<8;
    clen+=*buf++;

#ifdef USE_SUNCERT


    if (nsid==1 && ctype==1) {

	cdata=enlink_readudp(base);
	if (!cdata) return NULL;
	if (enlink_validate(mkid,len,cdata)) return NULL;
	res=db_cert_init();
	res->begins=enlink_getbefore(cdata);
        res->expires=enlink_getafter(cdata);
	if (res->p) int_drop(res->p);
	if (res->g) int_drop(res->g);
	if (res->ki) int_drop(res->ki);
        res->g=enlink_getg(cdata);
        res->p=enlink_getp(cdata);
        res->ki=enlink_geti(cdata);
	res->rns=nsid;
	res->cert_data=cdata;
        res->rmk=malloc(len);
        memcpy(res->rmk,mkid,len);
        db_unique_disk_save(&res,inet_ntoa(from),0);
        return res;
    }

#endif



    if (ctype != 4 || nsid != 8) {
#if 1
      printf("CERT add_cert unknown CERT-TYPE or NSID.\n");
#endif
        /* Huh? What? No comprende! */
        return NULL;
    }
    res=db_cert_init();
    res->rns=8; 
    t=buf;
    res->begins=*t++<<24;
    res->begins+=*t++<<16;
    res->begins+=*t++<<8;
    res->begins+=*t++;
    res->expires=*t++<<24;
    res->expires+=*t++<<16;
    res->expires+=*t++<<8;
    res->expires+=*t++;
    l=*t++<<8;
    l+=*t++;
    hex=malloc(2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->p) int_init(&(res->p));
    int_set_str(res->p,hex,16);
    l=*t++<<8;
    l+=*t++;
    hex=realloc(hex,2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->g) int_init(&(res->g));
    int_set_str(res->g,hex,16);
    l=*t++<<8;
    l+=*t++;
    hex=realloc(hex,2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->ki) int_init(&(res->ki));
    int_set_str(res->ki,hex,16);
    free(hex);
    if (t-buf!=clen) {
#if 1
        printf("CERT add_cert incoherent length.\n");
#endif
        db_cert_drop(res);
        return NULL;
    }
    res->rmk=malloc(16);
/*printf("%s",int_extract_hex(res->ki));*/
    cert_MD5Init(&ctxt);
    cert_MD5Update(&ctxt,buf,t-buf);
    cert_MD5Final(res->rmk,&ctxt);

    db_unique_disk_save(&res,inet_ntoa(from),0);

    return res;

}

void udp_process(int sock)
{
    extern void do_discovered(u_char nsid, u_char *mkid, dh_data *remote, 
    								u_long ip);
    dh_data *response=NULL;
    u_char buf[20000], *t;
    struct sockaddr_in addr;
    int len, addrlen=sizeof(addr);
    int q_nr,r_nr,nr;

    len=recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *) &addr, 
								&addrlen);
#if 0
    printf("CERT recvfrom(%s) ACT=%d STAT=%d NAMES=%d CERTS=%d\n",
           inet_ntoa(addr.sin_addr),*buf&3,buf[1]&127,buf[2],buf[3]);
#endif
    if (len<=0 || len <4) return;

    /* Simply return if ACTION 2 (response) and STATUS indicates an error */
    if (((*buf & 3) == 2) && (buf[1] & 127))
      return;

#ifdef UBS_CHIPCARD
    if ((*buf&3)==3)
    {
      ubs_process(sock, buf, len, &addr);
    }
#endif

#if 0
invalid code
    if ((*buf&15)==1 && buf[2]) {
      int i;
      printf("     MKID=");
      for(i=0;i<mkid_size[*(buf+2)];i++) printf("%02X", buf[4+i]);
      printf("\n");
    }
#endif

    t=buf;
    if ((*t & 0xf0) != 0x10) return; /* Version 1 only... */

    if (*t & 12) {
	printf("CERT can not handle cookies yet.\n");
	return;
    }

    t=buf+2;
    q_nr=*t++;
    r_nr=*t++;

    assert(q_nr <= 1 && r_nr <= 1);

    nr=q_nr;

    while(nr--) {
	/* this loop handles the requests. 
	 * current format is Name Type 
	 *			SKIP=1
	 *		     Name Length (8bit)
	 *		     Name 
	 *			NSID (8bit), MKID
	 */
	if (*t != 1) {
	    printf("CERT cert_process name type is not SKIP.\n");
	    /* really should return an error packet here! */
            t += 2+t[1];
	    continue;
	}
#ifdef USE_SUNCERT
        if (*(t+2)==1) {
            dh_data *find=db_public_find_load(NULL,*(t+2),t+3);
	    if (find == (dh_data *)(-1)) 
		find=db_public_find_load(find,*(t+2),t+3);
            if (find==NULL || find->cert_data==NULL) {
		*buf=(1<<4)|2;
		*(buf+1)=2;
		*(buf+2)=0;
		*(buf+3)=1;
                t += 2+t[1];
		*t++=0;
		*t++=0;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),
							*buf&3,buf[1],buf[3]);
#endif
		sendto(sock, buf, t-buf, 0,(struct sockaddr *) &addr, addrlen);
	        return; /* we forget the cert that was provided !!! ugly XXX*/
            } else {
	        int slen=enlink_saveudp(t, find->cert_data);
		if (!slen) return;
                *(t+2)=1;
		*buf=(1<<4)|2;
		*(buf+1)=0;
		*(buf+2)=0;
		*(buf+3)=1;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),
							*buf&3,buf[1],buf[3]);
#endif
		sendto(sock, buf, t-buf+slen, 0,(struct sockaddr *) &addr, 
								addrlen);
	        return; /* we forget the cert that was provided !!! ugly XXX*/
            } 

        } else if (*(t+2)==8) {
#else
        if (*(t+2)==8) {
#endif
            dh_data *find=db_public_find_load(NULL,*(t+2),t+3);
	    if (find == (dh_data *)(-1)) 
		find=db_public_find_load(find,*(t+2),t+3);
            if (find==NULL) {
		*buf=(1<<4)|2;
		*(buf+1)=2;
		*(buf+2)=0;
		*(buf+3)=1;
                t += 2+t[1];
		*t++=0;
		*t++=0;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),
							*buf&3,buf[1],buf[3]);
#endif
		sendto(sock, buf, t-buf, 0,(struct sockaddr *) &addr, addrlen);
	        return; /* we forget the cert that was provided !!! ugly XXX*/
            } else {
		u_char *raw;
                u_char *s;
                int l;

		/* send that bloody nsid8 cert */
		*buf=(1<<4)|2;
		*(buf+1)=0;
		*(buf+2)=0;
		*(buf+3)=1;
                t += 2+t[1];

                /* there might be more than one -- ignore for now */
                *t++=4; /* the only we know for now! */
                s=t;
                t+=2;
                /**t++=find->rns?find->rns:DEFAULT_NSID;*/
                l=find->begins;
		*t++=(l>>24)&0xff;
		*t++=(l>>16)&0xff;
		*t++=(l>>8)&0xff;
                *t++=l&0xff;
                l=find->expires;
		*t++=(l>>24)&0xff;
		*t++=(l>>16)&0xff;
		*t++=(l>>8)&0xff;
                *t++=l&0xff;
                l=int_extract_raw(&raw,find->p,0); assert(l<65536);
                *t++=l>>8;
                *t++=l&0xff;
                memcpy(t,raw,l);
                t+=l;
                free(raw);
                /* from now on, alignment is not guaranteed */
                l=int_extract_raw(&raw,find->g,0); assert(l<65536);
                *t++=l>>8;
                *t++=l&0xff;
                memcpy(t,raw,l);
                t+=l;
                free(raw);
                l=int_extract_raw(&raw,find->ki,0); assert(l<65536);
                *t++=l>>8;
                *t++=l&0xff;
                memcpy(t,raw,l);
                t+=l;
                free(raw);
                l=t-s-2; assert(l<65536);
                *s++=l>>8;
                *s=l&0xff;
#if 0
                printf("CERT sendto(%s) %d %d %d %d\n",inet_ntoa(addr.sin_addr),*buf&15,buf[1],buf[2],buf[3]);
#endif
                sendto(sock, buf, t-buf ,0 , (struct sockaddr *) &addr,
                                                                    addrlen);

	        return; /* we forget the cert that was provided !!! ugly XXX*/
	    }
        } else t += 2+t[1];
    }

    nr=r_nr;

    while(nr--) {
        dh_data *last;

        /* response */
        /* if positive response, find/create remote certificate and dh_data,
         * if the incoming data corresponds to an already existing cert,
         * it may be that it just contains new certificate data, which should
         * be appended. if different public value certificate with same name 
         * already exists, store the new one nevertheless, should later check 
         * validity times, and refuse the new one if overlap 
         * (or remove old one)
         * pass it with nsid, mkid, ip to do_discovered
         * if failure, just pass remote=NULL to do_discovered
         */
        last=add_cert(t,addr.sin_addr);
        /* old code 
	    if (last && last->rns == *(buf+2) && !memcmp(last->rmk,buf+4,
        						mkid_size[last->rns])) 
         */
        /* the memcmp also took care of cheating nsid8 replies, as saving */
        /* happens under true name anyway */

        if (last) response=last;

	/* XXX increment t here ! */
    }
    t=buf;
    if (r_nr) {
        /* positive OR negative response for do_discovered */
	/* for now, NO negative do_discovered do NOT happen XXX */
      if (response) 
        do_discovered(response->rns, response->rmk, response, 
						addr.sin_addr.s_addr);
    }
}


/* to ask about certificates. includes own public value if available */
void udp_request(u_long addr, u_char ons, u_char *omk, u_char rns, u_char *rmk)
{
    u_char buf[20000], *t;
    int l;
    dh_data *find;
    struct sockaddr_in iaddr;

    find=db_public_find_load(NULL,ons,omk);
    if (find == (dh_data *)(-1)) 
        find=db_public_find_load(find,ons,omk);
    t=buf;
    *t++=(1<<4)|1;
    *t++=0;
    *t++=1;
    *t++=0;

    *t++=1;
    *t++=mkid_size[rns]+1;
    *t++=rns;
    memcpy(t,rmk,mkid_size[rns]);
    t+=mkid_size[rns];

    memset(&iaddr,0,sizeof (iaddr));
    iaddr.sin_family = AF_INET;
    iaddr.sin_port = htons(RECV_PORT);
    iaddr.sin_addr.s_addr = addr; /* already network format */
#if 0
    printf("CERT sendto(%s) %d %d\n",inet_ntoa(iaddr.sin_addr),*buf&15,buf[1]);
#endif
    l=sendto(send_socket,buf,t-buf,0,(struct sockaddr *)&iaddr,sizeof(iaddr));
}
